@@ -43,7 +43,8 @@ class Worker:
         self.worker_name = worker_name
         if 'LtbCloudComponentsStorage' in os.environ:
             self.ltb_connection_string = os.environ['LtbCloudComponentsStorage']
-        self.processing_timeout = 120
+        else:
+            self.ltb_connection_string = None
         self.segment_limit = segment_limit
         self.batch_size = batch_size
 
@@ -93,71 +94,106 @@ class Worker:
             self.table_client.update_entity(state_table, state_row)
 
     def process_message(self, request_id, func, blob_func=None):
-        guid, request_hash = self._decode_id(request_id)
-        request_entity = self.table_client.get_entity(self.metadata_table_name, guid, request_hash)
         try:
-            self._process_request_body(request_entity, func)
-            self._process_blob(request_entity, blob_func)
+            guid, request_hash = self._decode_id(request_id)
+            request_entity = self.table_client.get_entity(self.metadata_table_name, guid, request_hash)
+            if 'Request' in request_entity:
+                self._process_request_body(request_entity, func)
+            if 'BlobName' in request_entity:
+                self._process_blob(request_id, request_entity, blob_func)
         except Exception as e:
-            if 'LtbCloudComponentsStorage' in os.environ:
-                track = traceback.format_exc()
-                function_failure_entity = {"PartitionKey": self.worker_name, "RowKey": request_id, "FailureMessage": str(e), "StackTrace": track}
-                TableService(connection_string=self.ltb_connection_string).insert_entity(failure_table, function_failure_entity)
+            if self.ltb_connection_string is not None:
+                self._record_failure(request_id, e)
             else:
                 raise
 
-    def process_blob(self, request_id, func):
-        guid, request_hash = self._decode_id(request_id)
-        request_entity = self.table_client.get_entity(self.metadata_table_name, guid, request_hash)
+    def _process_blob(self, request_id, request_entity, func):
+        blob_name, params, offset, retry_count, batch_count = self._read_metadata(request_entity)
         try:
-            blob_name = request_entity['BlobName']
-            if 'Parameters' in request_entity:
-                params = json.loads(request_entity['Parameters'])
+            result, whole_blob_read = self._process_blob_by_segments(blob_name, func, offset, params)
+        except Exception as e:
+            if retry_count == 3:
+                self._record_failure(request_id, e)
             else:
-                params = None
-
-            offset = 0
-            if 'Offset' in request_entity:
-                offset = request_entity['Offset'].value
-
-            in_blob = BlobClient.from_connection_string(self.ltb_connection_string, container_name=input_container, blob_name=blob_name)
-            result, batch = [], []
-            segment_id = -1
-            whole_blob_read = True
-            for source_segment in self._read_blob_by_segments(in_blob):
-                segment_id += 1
-                if segment_id < offset:
-                    continue
-                if len(batch) == self.batch_size:
-                    result.extend(func(batch, params))
-                    batch = []
-                else:
-                    batch.append(source_segment)
-
-                if segment_id - offset + 1 >= self.segment_limit:
-                    whole_blob_read = False
-                    break
-
-            if len(batch) > 0:
-                result.extend(func(batch, params))
-
-            serialized_result = "\n".join(json.dumps(r) for r in result)
-            out_blob = BlobClient.from_connection_string(self.ltb_connection_string, container_name=output_container, blob_name="%s/%s" % (self.worker_name, blob_name))
-            if offset == 0:
-                out_blob.upload_blob(serialized_result, timeout=600)
-            else:
-                out_blob.append_block("\n" + serialized_result)
-            
-            if not whole_blob_read:
-                request_entity['Offset'] = segment_id + 1
+                request_entity['RetryCount'] = retry_count + 1
                 self.table_client.update_entity(self.metadata_table_name, request_entity)
                 self._enqueue(request_id)
+            return
+            
+        serialized_result = "\n".join(json.dumps(r) for r in result)
+        if not whole_blob_read:
+            out_blob = self._get_output_blob("%s/%s_%d" % (self.worker_name, blob_name, batch_count))
+            out_blob.upload_blob(serialized_result, timeout=1200)
                 
-        except Exception as e:
-            track = traceback.format_exc()
-            function_failure_entity = {"PartitionKey": self.worker_name, "RowKey": request_id, "FailureMessage": str(e), "StackTrace": track}
-            TableService(connection_string=self.ltb_connection_string).insert_entity(failure_table, function_failure_entity)
+            request_entity['Offset'] = offset + len(result)
+            request_entity['Batch'] = batch_count + 1
+            self.table_client.update_entity(self.metadata_table_name, request_entity)
+            self._enqueue(request_id)
+        else:
+            out_blob = self._get_output_blob("%s/%s" % (self.worker_name, blob_name))
+            if offset == 0:
+                out_blob.upload_blob(serialized_result, timeout=1200)
+            else:     
+                content = []
+                for batch in range(batch_count):
+                    in_blob = self._get_output_blob("%s/%s_%d" % (self.worker_name, blob_name, batch))
+                    blob_content = in_blob.download_blob().readall().decode()
+                    content.append(blob_content)
+                content.append(serialized_result)
+                out_blob.upload_blob("\n".join(content), timeout=1200)
 
+    def _get_input_blob(self, blob_name):
+        return BlobClient.from_connection_string(self.ltb_connection_string, container_name=input_container, blob_name=blob_name)
+
+    def _get_output_blob(self, blob_name):
+        return BlobClient.from_connection_string(self.ltb_connection_string, container_name=output_container, blob_name=blob_name)
+
+    def _read_metadata(self, request_entity):
+        blob_name = request_entity['BlobName']
+        params = None
+        if 'Parameters' in request_entity:
+            params = json.loads(request_entity['Parameters'])
+        offset = 0
+        if 'Offset' in request_entity:
+            offset = request_entity['Offset']
+        retry_count = 0
+        if 'RetryCount' in request_entity:
+            retry_count = request_entity['RetryCount']
+        batch_count = 0
+        if 'Batch' in request_entity:
+            batch_count = request_entity['Batch']
+        return blob_name, params, offset, retry_count, batch_count
+
+    def _process_blob_by_segments(self, blob_name, func, offset, params):
+        result = []
+        in_blob = self._get_input_blob(blob_name)
+        batch = []
+        segment_id = -1
+        whole_blob_read = True
+        for source_segment in self._read_blob_by_segments(in_blob):
+            segment_id += 1
+            if segment_id < offset:
+                continue
+
+            batch.append(source_segment)
+            if len(batch) == self.batch_size:
+                result.extend(func(batch, params))
+                batch = []
+
+            if segment_id - offset + 1 >= self.segment_limit:
+                whole_blob_read = False
+                break
+
+        if len(batch) > 0:
+            result.extend(func(batch, params))
+            
+        return result, whole_blob_read
+
+    def _record_failure(self, request_id, e: Exception):
+        track = traceback.format_exc()
+        function_failure_entity = {"PartitionKey": self.worker_name, "RowKey": request_id, "FailureMessage": str(e), "StackTrace": track}
+        TableService(connection_string=self.ltb_connection_string).insert_entity(failure_table, function_failure_entity)
+        
     def get_result(self, request_id):
         guid, request_hash = self._decode_id(request_id)
         entity = self.table_client.get_entity(self.metadata_table_name, guid, request_hash)
@@ -191,26 +227,6 @@ class Worker:
         request_entity['Result'] = json.dumps(response)
         self.table_client.update_entity(self.metadata_table_name, request_entity)
 
-    def _process_blob(self, request_entity, func):
-        if 'BlobName' not in request_entity:
-            return
-
-        blob_name = request_entity['BlobName']
-        if 'Parameters' in request_entity:
-            params = json.loads(request_entity['Parameters'])
-        else:
-            params = None
-
-        in_blob = BlobClient.from_connection_string(self.ltb_connection_string, container_name=input_container, blob_name=blob_name)
-        out_blob = BlobClient.from_connection_string(self.ltb_connection_string, container_name=output_container, blob_name="%s/%s" % (self.worker_name, blob_name))
-
-        result = []
-        for source_segments in self._read_blob_by_chunks(in_blob):
-            dnts = func(source_segments, params)
-            result.extend(dnts)
-        serialized_result = "\n".join(json.dumps(dnts) for dnts in result)
-        out_blob.upload_blob(serialized_result, timeout=600)
-
     def _wake_up(self):
         if self.should_wake_up:
             self.wake_up_queue.send_message("d2FrZSB1cA==")
